def define_tpu_strategy(mixed_precision = False, xla_accelerate = False):
	try:
		tpu = tf.distribute.cluster_resolver.TPUClusterResolver()
		print('Running on TPU ', tpu.master())
	
	except ValueError:
		print("TPU not activated. Please check the settings. Settings -> Accelerator -> TPU v3-8\n")
		tpu = None
	
	
	if tpu:
		tf.config.experimental_connect_to_cluster(tpu)
		tf.tpu.experimental.initialize_tpu_system(tpu)
		strategy = tf.distribute.experimental.TPUStrategy(tpu)
	else:
		strategy = tf.distribute.get_strategy()

	if mixed_precision:
		from tensorflow.keras.mixed_precision import experimental as mixed_precision 
		if tpu: 
			policy = tf.keras.mixed_precision.experimental.Policy('mixed_bfloat16')
		else:
			policy = tf.keras.mixed_precision.experimental.Policy('mixed_float16')
		mixed_precision.set_policy(policy)
		print("Mixed precision enabled")
	else:
		print("Mixed precision disabled")
	if xla_accelerate:
		tf.config.optimizer.set_jit(True)
		print("Accelerated Linear Algebra enabled")
	else:
		print("Accelerated Linear Algebra disabled")
	
	return strategy, tpu


############### DATASET SETUP  ###################


def decode_image(image_data, IMAGE_SIZE):
	
	image = tf.io.decode_raw(image_data, tf.uint8)
	
	image = tf.cast(image, tf.float32) / 255.0
	

	# how would I obtain Image_Size 
	image = tf.reshape(image, [*IMAGE_SIZE, 3])
	return image

def read_labeled_tfrecord(example):
	
	global dictionary_labeled
	global IMAGE_SIZE
	global tmp_d_l
	global tmp_img_sz
	
	tmp_d_l = dictionary_labeled
	tmp_img_sz = IMAGE_SIZE
	
	if dictionary_labeled is None:
		print("Enter the dictionary of Labeled_TFREC_FORMAT\n")
		dictionary = eval(input())
	else:
		dictionary = eval(dictionary_labeled)
	
	if IMAGE_SIZE is None:
		print("Enter Image Size; Example Format ->  192,192  (w/o brackets)\n\n")
		IMAGE_SIZE = input()
		
	IMAGE_SIZE = [int(IMAGE_SIZE.split(',')[0]), int(IMAGE_SIZE.split(',')[1])]
	example = tf.io.parse_single_example(example, dictionary)
	image = decode_image(example[list(dictionary.keys())[0]], IMAGE_SIZE)
	label = tf.cast(example[list(dictionary.keys())[1]], tf.int32)
	
	dictionary_labeled = tmp_d_l
	IMAGE_SIZE = tmp_img_sz
	
	return image, label

def read_unlabeled_tfrecord(example):
	
	global dictionary_unlabeled
	global IMAGE_SIZE
	global tmp_d_ul
	global tmp_img_sz
	
	tmp_d_ul = dictionary_unlabeled
	tmp_img_sz = IMAGE_SIZE
	
	
	if dictionary_unlabeled is None:
		print("Enter the dictionary of the UnLabeled_TFREC_FORMAT\n")
		dictionary = eval(input())
	else:
		dictionary = eval(dictionary_unlabeled)
	
	if IMAGE_SIZE is None:
		print("Enter Image Size; Example Format ->  192,192  (w/o brackets)\n\n")
		IMAGE_SIZE = input()
	IMAGE_SIZE = [int(IMAGE_SIZE.split(',')[0]), int(IMAGE_SIZE.split(',')[1])]
	example = tf.io.parse_single_example(example, dictionary)
	image = decode_image(example[list(dictionary.keys())[0]], IMAGE_SIZE)
	idnum = example[list(dictionary.keys())[1]]
	dictionary_unlabeled = tmp_d_ul
	IMAGE_SIZE = tmp_img_sz
	
	return image, idnum


def load_dataset(filenames, labeled = True, ordered = False):
	ignore_order = tf.data.Options()
	if not ordered:
		ignore_order.experimental_deterministic = False
		
	dataset = tf.data.TFRecordDataset(filenames, num_parallel_reads = tf.data.experimental.AUTOTUNE) # can add AUTO if multiple files need to be read in a go.
	dataset = dataset.with_options(ignore_order)
	dataset = dataset.map(read_labeled_tfrecord if labeled else read_unlabeled_tfrecord)
	return dataset


def get_training_dataset(GCS_DS_PATH, train_tfrec_path, BATCH_SIZE, augmentation = False):
	
	if not augmentation:
		print(""" Make Sure to
		Define how to read LABELED tfrecord data as per the LABELED_TFRECORD_FORMAT
		Rest of the helper functions are implemented.
		""")
	dataset = load_dataset(tf.io.gfile.glob(GCS_DS_PATH  + train_tfrec_path), labeled = True)
	if augmentation:
		from quick_ml.augments import augmentations
		dataset = dataset.map(augmentations, num_parallel_calls = tf.data.experimental.AUTOTUNE)
	cnt = 0
	for data in dataset:
		cnt += 1
	if not augmentation:
		print(f"Loaded {train_tfrec_path} with {cnt} examples")
	dataset = dataset.repeat()
	dataset = dataset.shuffle(2048)
	dataset = dataset.batch(BATCH_SIZE)
	dataset = dataset.prefetch(tf.data.experimental.AUTOTUNE)
	return dataset


def get_validation_dataset(GCS_DS_PATH,val_tfrec_path, BATCH_SIZE , augmentation = False):
	
	
	if not augmentation:
		print(""" Make Sure to
		Define how to read LABELED tfrecord data as per the LABELED_TFRECORD_FORMAT
		Rest of the helper functions are implemented.
		""")
	
	dataset = load_dataset(tf.io.gfile.glob(GCS_DS_PATH + val_tfrec_path), labeled = True, ordered = False)
	
	if augmentation:
		from quick_ml.augments import augmentations
		dataset = dataset.map(augmentations, num_parallel_calls = tf.data.experimental.AUTOTUNE)
	cnt = 0
	for data in dataset:
		cnt += 1
	if not augmentation:
		print(f"Loaded {val_tfrec_path} with {cnt} examples")
	dataset = dataset.batch(BATCH_SIZE)
	dataset = dataset.cache()
	dataset = dataset.prefetch(tf.data.experimental.AUTOTUNE)
	
	return dataset


def get_test_dataset(GCS_DS_PATH, test_tfrec_path, BATCH_SIZE):
	
	print(""" Make Sure to
	Define how to read UNLABELED tfrecord data as per the UNLABELED_TFRECORD_FORMAT
	Rest of the helper functions are implemented.
	""")

	dataset = load_dataset(tf.io.gfile.glob(GCS_DS_PATH + test_tfrec_path), labeled = False, ordered = True)
	cnt = 0
	for data in dataset:
		cnt += 1
	print(f"Loaded {test_tfrec_path} with {cnt} examples")
	dataset = dataset.batch(BATCH_SIZE)
	return dataset


def count_num_examples(tfrecordfile, labeled = True):
	
	tfdataset = load_dataset(tfrecordfile, labeled)
	cnt = 0
	for data in tfdataset:
		cnt += 1
	print(f"{tfrecordfile} contains {cnt} Examples")

def get_labeled_tfrecord_format(dict, img_size):
	global dictionary_labeled
	global IMAGE_SIZE
	
	dictionary_labeled = dict
	IMAGE_SIZE = img_size
	
def get_unlabeled_tfrecord_format(dict, img_size):
	global dictionary_unlabeled
	global IMAGE_SIZE
	
	dictionary_unlabeled = dict
	IMAGE_SIZE = img_size

## later	
def set_parameters(epochs, batch_size_base, GCS_PATH):
	from kaggle_datasets import KaggleDatasets
	GCS_DS_PATH = KaggleDatasets().get_gcs_path(GCS_PATH)
	print(f'GCS_DS_PATH ->  {GCS_DS_PATH}')
	
	STEPS_PER_EPOCH = None

if __name__ != "__main__":
	import tensorflow as tf
	if tf.__version__ != '2.4.0':
		print("Error! Tensorflow version mismatch")
		
	



## script to remove class imbalances
## Use this script if there is any imbalance among the classes else don't

## Direct Mode (yet to be written)
# def upsample():
#     pass

# def downsample():


#     pass


# def check_transformations(func):
#     acceptable = ['random_crop', 'flip_left_right', 'flip_up_down',
#     'random_hue', 'random_contrast', 'brightness', 
#     'random_saturation', 'zoom', 'rotation']
    
#     def inner(lst):

#         for i in lst:
#             if i not in acceptable:
#                 raise Exception(f"Error! Couldn't understand the transformation -> {i}. \n Check help(define_transformations) to know supported transformations.")

#         func(lst)

#     return inner



# @check_transformations
# def define_transformations(lst):
#     """
#     Supported Transformations -> 
#     'random_crop', 'flip_left_right', 'flip_up_down',
#     'random_hue', 'random_contrast', 'brightness', 
#     'random_saturation', 'zoom', 'rotation', 'mixup', 'cutmix'
#     """
#     global transformations_list

#     if len(lst) > 0:
#         transformations_list = lst
#     else:
#         raise Exception("Error! No transformation defined...")





### Interactive Mode
class Class_Balancer():
    def __init__(self, training_data_path, output_path = None, use_multiprocessing = True):
        self.balanced = False
        self.max_class_samples = None
        self.training_data_path = training_data_path
        self.classes = None
        self.output_path = output_path
        self.multiprocessing = use_multiprocessing

        if self.__check_imbalance(training_data_path):
            self.balanced = True

        if not self.balanced:  
            print("\nThe dataset is imbalanced. Please mention the set of transformations to be applied.")
            self.transformations_list = ['random_crop', 'flip_left_right', 'flip_up_down',
            'random_hue', 'random_contrast', 'random_brightness', 
            'random_saturation', 'random_zoom', 'random_rotation']#, 'mixup', 'cutmix']
            self.bool_transformations = ['flip_left_right', 'flip_up_down'] # mixup, cutmix
            #self.input_values = {'random_crop' : [crop_height, crop_width, channels], 'random_hue' : [0, 0.5), 'random_contrast' : [lower, upper], 'brightness' : [+max_delta], 'random_saturation' : [lower, upper], 'random_zoom' : [(height, width), fill_mode], 'random_rotation' : [rg, fill_mode]}
            #self.values_constraints = {'random_crop' : }
            self.input_value_parameters = {'random_crop' : ['crop_height', 'crop_width', 'channels'], 'random_hue' : ['max_delta'], 'random_contrast' : ['lower', 'upper'], 'random_brightness' : ['max_delta'], 'random_saturation' : ['lower', 'upper'], 'random_zoom' : ['height', 'width', 'fill_mode'], 'random_rotation' : ['rg', 'fill_mode']}
            self.user_transformations = {}
            self.user_selections = []

            for i in self.transformations_list:
                print('\n', i, " : ")
                response = input('Please enter in yes/no \n')
                self.user_transformations[i] = []
                self.user_transformations[i].append(True if response.lower()[0] == 'y' else False)
                if self.user_transformations[i][0] and (i not in self.bool_transformations):
                    for param in self.input_value_parameters[i]:
                        self.user_transformations[i].append(input(f"\nPlease enter the value for {i} parameter {param} : \n"))

            for item in self.user_transformations.items():
                if item[1][0]:
                    self.user_selections.append(item[0])

            if not any(list(self.user_transformations.values())):
                raise Exception("Please mention atleast one of the transformations to be applied...")


    @staticmethod
    def augment_class():
        pass

    def __check_imbalance(self, training_data_path):
        """
        Helps you to visualize the class imbalance by plotting the number of data points
        per class.
        """
        classes = os.listdir(training_data_path)
        #self.classes = classes
        class_dist = {}
        for class_ in classes:
            class_dist[class_] = len(glob(training_data_path + '/' + class_ + "/*.jpg"))
        
        self.classes = class_dist
        self.max_class_samples = max(list(class_dist.values()))
        #print(class_dist)
        plt.figure(figsize = (20,20))
        sns.barplot(x = list(class_dist.keys()), y = list(class_dist.values()))
        plt.xticks(rotation = 90)
        plt.show()
        keys = list(class_dist.keys())
        set_values = set()
        for k in keys:
            set_values.add(class_dist[k])
        return len(set_values) == 1

    def __transformations_func(self, img, transform):
        if transform == "random_crop":
            crop1 = int(self.user_transformations[transform][1])
            crop2 = int(self.user_transformations[transform][2])
            crop3 = int(self.user_transformations[transform][3])
            if img.shape[0] < crop1 :
                crop1 = img.shape[0]
            if img.shape[1] < crop2 : 
                crop2 = img.shape[1]
            
            return tf.image.random_crop(img, [crop1,crop2,crop3] )
            #return tf.image.random_crop(img, [int(self.user_transformations[transform][1]), int(self.user_transformations[transform][2]), int(self.user_transformations[transform][3])])
        elif transform == "random_zoom":
            zoom1 = float(self.user_transformations[transform][1])
            zoom2 = float(self.user_transformations[transform][2])

            if img.shape[0] > zoom1:
                zoom1 = img.shape[0]
            if img.shape[1] > zoom2:
                zoom2 = img.shape[1]
            
            return tf.keras.preprocessing.image.random_zoom(img, (zoom1, zoom2), fill_mode = self.user_transformations[transform][3])
            
            #return tf.keras.preprocessing.image.random_zoom(img, (float(self.user_transformations[transform][1]), float(self.user_transformations[transform][2])), fill_mode = self.user_transformations[transform][3])
        elif transform == "flip_left_right":
            return tf.image.flip_left_right(img)
        elif transform == "flip_up_down":
            return tf.image.flip_up_down(img)
        elif transform == "random_brightness":
            return tf.image.random_brightness(img, float(self.user_transformations[transform][1]))
        elif transform == "random_contrast":
            return tf.image.random_contrast(img, float(self.user_transformations[transform][1]), float(self.user_transformations[transform][2]))
        elif transform == "random_hue":
            return tf.image.random_hue(img, float(self.user_transformations[transform][1]))
        elif transform == "random_saturation":
            return tf.image.random_saturation(img, float(self.user_transformations[transform][1]), float(self.user_transformations[transform][2]))
        elif transform == "random_rotation":
            return tf.keras.preprocessing.image.random_rotation(img, float(self.user_transformations[transform][1]), fill_mode = self.user_transformations[transform][2], row_axis = 0, col_axis = 1, channel_axis = 2,)
      

    def __augment_multi_classes(self, class_lst, path, verbose):
        for class_ in class_lst:
            self.__augment_class_samples(class_, path, verbose)

    def __augment_class_samples(self, class_, path, verbose):
        diff = self.max_class_samples - len(glob( path + '/' + class_ + '/*.jpg'))
        augmentations_to_be_applied = np.random.choice(self.user_selections, replace = True, size = diff)
        indices = np.random.choice( range(len(glob( path + '/' + class_ + "/*.jpg"))),replace = True, size = diff)
        images_ = glob( path + '/' + class_ + '/*.jpg')
        
        #if verbose:
        print(f"Augmenting for class {class_}")
        for cnt, i in enumerate(tqdm(augmentations_to_be_applied)):
            img = cv2.imread(images_[indices[cnt]])
            #print(img.shape)
            temp_img = self.__transformations_func(img, i)
            temp_img = np.array(temp_img)
            #temp_img.save(os.path.join(path , images_[cnt] + '.jpg'))
            if verbose:
                print(images_[indices[cnt]], " | transform | ", i)
            save_name = images_[indices[cnt]].split('//')[-1].split('.')[0] + "_" + str(cnt) + '.jpg'
            cv2.imwrite( save_name, temp_img)
    


    def __augment_class_numbers(class_, path, num_samples):
        "Method under making..."
        pass 

    def balance_classes(self, inplace = True, output_directory = None, verbose = True):
        save_path = None
        if inplace is False:
            assert output_directory is not None
            save_path = output_directory
        else:
            save_path = self.training_data_path


        mode = int(input("\nPlease specify the mode of class balancing... -> \n1. Class Wise Balance \n2. Certain number of Samples Balance\n Please Choose between (1 or 2)\n"))
        if mode == 1:
          
            if self.multiprocessing:

                inp_classes = list(self.classes.keys())

                p1 = Process(target = self.__augment_multi_classes, args = (inp_classes[:len(inp_classes)//4], save_path,verbose))
                p2 = Process(target = self.__augment_multi_classes, args = (inp_classes[len(inp_classes)//4 : 2 * (len(inp_classes)//4)], save_path, verbose))
                p3 = Process(target = self.__augment_multi_classes, args = (inp_classes[2 * (len(inp_classes)//4) : 3 * (len(inp_classes) // 4)], save_path, verbose))
                p4 = Process(target = self.__augment_multi_classes, args = (inp_classes[3 * (len(inp_classes) // 4) : ], save_path, verbose))

                p1.start()
                p2.start()
                p3.start()
                p4.start()

                p1.join()
                p2.join()
                p3.join()
                p4.join()

            else:
                for class_ in list(self.classes.keys()):
                    self.__augment_class_samples(class_, save_path, verbose)

          
            print("Data Balancing complete...")        

        else:
            num_samples = int(input("\nPlease enter the number of samples (> max_samples in any class)\n"))
            assert num_samples > self.max_class_samples

            for class_ in list(self.classes.keys()):
                self.__augment_class_numbers(class_, save_path, num_samples)
            print("Data balancing complete...")
        self.balanced = True


    def csv_to_folder(self):
        print("Method under making...")
        pass

    def make_tfrecords(self):
        print("Method under making...")
        if self.balanced:
            pass
        else:
            raise Exception("Dataset path not yet balanced. Please balance the dataset by invoking balance_classes method of the class instance.")



if __name__ != "__main__":
    import tensorflow as tf 
    import os
    import seaborn as sns 
    import matplotlib.pyplot as plt
    import numpy as np
    import cv2
    from tqdm import tqdm
    from PIL import Image
    from glob import glob
    from multiprocessing import Process 
     
    sns.set_style('darkgrid')
    if tf.__version__ != '2.4.0':
        raise Exception("Tensorflow version mismatch!")

    global class_dist




class V2F():

    def __init__(self, output_folder):
        self.output_folder = output_folder



    def __get_fps(self, video_file):
        video = cv2.VideoCapture(video_file);

        (major_ver, minor_ver, subminor_ver) = (cv2.__version__).split('.')

        if int(major_ver)  < 3 :
            fps = video.get(cv2.cv.CV_CAP_PROP_FPS)
            print ("Frames per second using video.get(cv2.cv.CV_CAP_PROP_FPS): {0}".format(fps))
        else:
            fps = video.get(cv2.CAP_PROP_FPS)
            print ("Frames per second using video.get(cv2.CAP_PROP_FPS) : {0}".format(fps))

        video.release()

        return int(round(fps))



    def convert2frames(self, video_path, video_name, sparsify = True, frame_freq = None):
        assert os.path.isfile(video_path) == True
        
        if sparsify:
            #frame_freq = int(input("Please enter the save frequency for frames per second: \n"))
            assert frame_freq is not None 
            assert isinstance(frame_freq, int)

        cam = cv2.VideoCapture(video_path)

        if not os.path.exists(self.output_folder):
            os.makedirs(self.output_folder) 

        currentframe = 0
        min_time_counter = 1
        frame_counter = 0
        fps = self.__get_fps(video_path)
        intervals = random.sample(range(0, fps), frame_freq)
        random_intervals = []

        min_intervals = []
        for i in range(0, 60):
            for j in intervals:
                min_intervals.append(j + (fps*i))

        while(True):

            ret, frame = cam.read()

            if ret:
                frame_counter += 1

                name = f'{self.output_folder}/{video_name}/{str(min_time_counter)}/frame' + str(currentframe) + '.jpg'
                #print ('Creating...' + name)
                if not os.path.exists(self.output_folder + '/' + video_name + '/' + str(min_time_counter)):
                    os.makedirs(self.output_folder + '/' + video_name + '/' + str(min_time_counter))
  
                # writing the extracted images
                if sparsify == "yes" :
        
        
        
                    #if frame_counter % 60 == 0:
        
                    #if divisible(frame_counter, random_intervals):
                    #if frame_counter in random_intervals:
                    if frame_counter in min_intervals:
            
                        cv2.imwrite(name, frame)
                        print ('Creating...' + name)
                else:
                    cv2.imwrite(name, frame)
                    print ('Creating...' + name)


                # increasing counter so that it will
                # show how many frames are created
                currentframe += 1

                if frame_counter >= fps * 60:
                  #start_time = time.time()
                    min_time_counter += 1
                    frame_counter = 0

            else:
                break

        cam.release()
        cv2.destroyAllWindows()


    def extract_interval(self):
        pass 

    def extract_multi_intervals(self):
        pass
    

    def convert_batch2frames(self):
        pass

    def compress(self, output_file_name):

        zipf = zipfile.ZipFile(output_file_name, 'w', zipfile.ZIP_DEFLATED)
        self.__zipdir(self.output_folder, zipf)
        zipf.close()


    def __zipdir(self, path, ziph):
        # ziph is zipfile handle
        for root, dirs, files in os.walk(path):
            for file in files:
                ziph.write(os.path.join(root, file), 
                           os.path.relpath(os.path.join(root, file), 
                                           os.path.join(path, '..')))



if __name__ != "__main__":
    import os
    import cv2 
    import numpy as np 
    from tqdm import tqdm 
    from glob import glob
    import random
    import zipfile
    
      



import tensorflow as tf
if tf.__version__ == '2.4.0':
	print(f"Tensorflow imported successfully. Tensorflow version -> {tf.__version__}")
else:
	print("Tensorflow version mismatch!. Please check tensorflow isn't imported before installing quick_ml. Restart the session to fix the error.")

from quick_ml import tfrecords_maker
from quick_ml import begin_tpu
from quick_ml import visualize_and_check_data
from quick_ml import load_models_quick
from quick_ml import training_predictions
from quick_ml import predictions
from quick_ml import augments
from quick_ml import k_fold_training
#import quick_ml.data_preprocessing.remove_imbalance
__version__ = '1.3.9'



def grad_cam(input_model, image, cls, layer_name, H=320, W = 320):
	"""GradCAM method for visualizing input saliency."""

	y_c = input_model.output[0, cls]
	conv_output = input_model.get_layer(layer_name).output
	grads = K.gradients(y_c, conv_output)



	gradient_function = K.function([input_model.input], )





def compute_gradcam(model, img, image_dir, df, labels, selected_labels, layer_name = 'bn'):
	preprocessed_input = load_image(img, image_dir, df)

	predictions = model.predict(preprocessed_input)

	print("Loading original image...")
	plt.figure(figsize = (15,10))
	plt.subplot(151)
	plt.title("Original....")
	plt.axis("off")
	plt.imshow(load_image(img, image_dir, df, preprocess = False), cmap = 'grey')

	j = 1

	for i in range(len(labels)):

		if labels[i] in selected_labels:

			print(f"Generating gradcam for class {labels[i]}")
			gradcam = grad_cam(model, preprocessed_input, i, layer_name)
			plt.subplot(151 + j)
			plt.title(f"{labels[i]}: p={predictions[0][i]:.3f}")
            plt.axis('off')
            plt.imshow(load_image(img, image_dir, df, preprocess=False),
                       cmap='gray')
            plt.imshow(gradcam, cmap='jet', alpha=min(0.5, predictions[0][i]))
            j += 1



def get_roc_curve(labels, predicted_vals, generator):
	auc_roc_vals = []

	for i in range(len(labels)):
		try:
			pass	
		except Exception as e:
			print(f"Error in generating ROC Curve values for {labels[i]}")

	plt.show()
	return auc_roc_vals




def get_weighted_loss(pos_weights, neg_weights, epsilon = 1e-7):

	def weighted_loss(y_true, y_pred):

		loss = 0.0


		for i in range(len(pos_weights)):

			loss += -1 * K.mean((pos_weights[i] * y_true[:, i] * K.log(y_pred[:,i] + epsilon)))





if __name__ != "__main__":
	import random
	import cv2
	import matplotlib.pyplot as plt 
	from keras import backend as K 
	from keras.preprocessing import image 
	from sklearn.metrics import roc_auc_score, roc_curve 
	from tensorflow.compat.v1.logging import INFO, set_verbosity


### https://www.kaggle.com/rftexas/implementing-gradcam-with-keras-for-error-analysis


class train_k_fold_pred():

	def __init__(self, k, tpu, n_class, model_name, train_tfrecs_path, val_tfrecs_path, test_tfrecs_path, GCS_DS_PATH_labeled, GCS_DS_PATH_unlabeled, BATCH_SIZE, EPOCHS, classification_model = 'default', freeze = False, input_shape = [512,512,3], activation = 'softmax', weights = "imagenet", optimizer = "adam", loss = "sparse_categorical_crossentropy", metrics = "sparse_categorical_accuracy", callbacks = None, plot = False, verbose = 0):
		# the number of folds , k
		self.k = k
		self.tpu = tpu
		self.n_class = n_class
		self.model_name = model_name
		
		# for saving the models of each fold
		self.models = []
		# for saving model's folds history
		self.history = []
		# for saving the probabilities of predictions
		self.probabilities = []

		# obtain the training tfrecs path
		self.train_tfrecs_path = train_tfrecs_path
		# obtain the validation tfrecs path
		self.val_tfrecs_path = val_tfrecs_path
		# obtain the test tfrecs path
		self.test_tfrecs_path = test_tfrecs_path

		self.GCS_DS_PATH_labeled = GCS_DS_PATH_labeled
		self.GCS_DS_PATH_unlabeled = GCS_DS_PATH_unlabeled
		self.BATCH_SIZE = BATCH_SIZE
		self.EPOCHS = EPOCHS
		self.classification_model = classification_model
		self.freeze = freeze
		self.input_shape = input_shape
		self.activation = activation
		self.weights = weights
		self.optimizer = optimizer
		self.loss = loss
		self.metrics = metrics
		self.callbacks = callbacks
		self.plot = plot
		self.verbose = verbose
		
		from quick_ml.begin_tpu import get_test_dataset
		self.test_ds = get_test_dataset(self.GCS_DS_PATH_unlabeled, self.test_tfrecs_path, self.BATCH_SIZE)
		self.test_image_ds = self.test_ds.map(lambda image, idnum : image)
		self.NUM_TEST_IMAGES = 0
		for data in self.test_ds.unbatch():
			self.NUM_TEST_IMAGES += 1



	def train_k_fold(self):

		kfolds = KFold(self.k, shuffle = True, random_state = 21)

		TRAINING_FILENAMES = tf.io.gfile.glob(self.GCS_DS_PATH_labeled + self.train_tfrecs_path) + tf.io.gfile.glob(self.GCS_DS_PATH_labeled + self.val_tfrecs_path)

		from quick_ml.load_models_quick import create_model
		from quick_ml.begin_tpu import load_dataset
		
		
		df = pd.DataFrame(columns = ['Model_Fold_Number', 'Accuracy_top1', 'Accuracy_top3', 'Val_Accuracy_top1', 'Val_Accuracy_top3'])

		for fold, (trn_ind, val_ind) in enumerate(kfolds.split(TRAINING_FILENAMES)):
			print(f"Beginning with Model {self.model_name} : Fold Number -> {fold + 1}")
			train_dataset = load_dataset(list(pd.DataFrame({'TRAINING_FILENAMES' : TRAINING_FILENAMES}).loc[trn_ind]['TRAINING_FILENAMES']), labeled = True)
			val_dataset = load_dataset(list(pd.DataFrame({'TRAINING_FILENAMES' : TRAINING_FILENAMES}).loc[val_ind]['TRAINING_FILENAMES']), labeled = True, ordered = True)

			STEPS_PER_EPOCH = len(list(train_dataset)) // self.BATCH_SIZE

			train_dataset = train_dataset.repeat()
			train_dataset = train_dataset.shuffle(2048)
			train_dataset = train_dataset.batch(self.BATCH_SIZE)
			train_dataset = train_dataset.prefetch(tf.data.experimental.AUTOTUNE)

			val_dataset = val_dataset.batch(self.BATCH_SIZE)
			val_dataset = val_dataset.cache()
			val_dataset = val_dataset.prefetch(tf.data.experimental.AUTOTUNE)


			tf.tpu.experimental.initialize_tpu_system(self.tpu)
			strategy = tf.distribute.experimental.TPUStrategy(self.tpu)
			with strategy.scope():
				if self.classification_model != 'default':
					model = create_model(freeze = self.freeze, input_shape = self.input_shape, activation = self.activation, weights= self.weights, optimizer = self.optimizer, loss = self.loss, metrics = self.metrics, classes = self.n_class, model_name = self.model_name, classification_model = self.classification_model)
				else:
					model = create_model(freeze= self.freeze, input_shape = self.input_shape, activation = self.activation, weights = self.weights, optimizer = self.optimizer, loss = self.loss, metrics = self.metrics, classes = self.n_class, model_name = self.model_name)

			history = model.fit(train_dataset, epochs = self.EPOCHS, steps_per_epoch = STEPS_PER_EPOCH, validation_data = val_dataset, verbose = self.verbose)

			self.probabilities.append(model.predict(self.test_image_ds))

			tf.keras.backend.clear_session()
			tf.compat.v1.reset_default_graph()
			del model
			gc.collect()

			df = df.append(pd.DataFrame([[self.model_name + f'_fold_{fold+1}', history.history[self.metrics][-1], np.mean(history.history[self.metrics][-3:]), history.history['val_' + self.metrics][-1], np.mean(history.history['val_' + self.metrics][-3:])]], columns = ['Model_Fold_Number', 'Accuracy_top1', 'Accuracy_top3', 'Val_Accuracy_top1', 'Val_Accuracy_top3']), ignore_index = True)
			print(f'Done with the model fold -> {self.model_name}_fold_{fold+1}')
		return df

	def obtain_predictions(self, weights = None):

		if weights is not None:
			if len(weights) == len(self.probabilities):
				self.probabilities = np.average(self.probabilities, weights = weights, axis = 0)
			else:
				raise Exception("Length of weights array is not equal to the number of folds... Error!")
		else:
			self.probabilities = np.average(self.probabilities, axis = 0)
		
					
		test_ids_ds = self.test_ds.map(lambda image, idnum : idnum).unbatch()
		test_ids = next(iter(test_ids_ds.batch(self.NUM_TEST_IMAGES))).numpy().astype('U')
		print("Generating predictions output file as predictions.csv")
		
		
		if self.n_class == 1:
			predictions = np.array([1 if i >= 0.5 else 0 for i in self.probabilities])
			predictions = np.reshape(predictions, newshape = (self.NUM_TEST_IMAGES,))
		else:
			predictions = np.argmax(self.probabilities, axis = -1)
		np.savetxt('predictions.csv', np.rec.fromarrays([test_ids, predictions]),
			fmt = ['%s', '%d'], delimiter = ',', header = 'id,label', comments = '')


if __name__ != '__main__':
	import numpy as np
	import pandas as pd
	from sklearn.model_selection import KFold
	import os
	import gc
	import tensorflow as tf
	if tf.__version__ != '2.4.0':
		raise Exception("Error! Tensorflow version mismatch!...")
	



## function for obtaining predictions through GCS_DS_PATH
def get_predictions_tfrec(GCS_DS_PATH, test_tfrec_path, BATCH_SIZE, model, output_filename):
	import numpy as np
	from quick_ml.begin_tpu import get_test_dataset
	testdata = get_test_dataset(GCS_DS_PATH, test_tfrec_path, BATCH_SIZE)
	images = testdata.map(lambda image, idnum : image)
	ids = testdata.map(lambda image, idnum : idnum).unbatch()

	NUM_TEST_IMAGES = 0
	for data in testdata.unbatch():
		NUM_TEST_IMAGES += 1

	#probabilities = model.predict(images)
	#predictions = np.argmax(probabilities,axis = -1)
	predictions = model.predict_classes(images)
	predictions = np.reshape(predictions, newshape = (NUM_TEST_IMAGES,))
	print("Generating Output File...")
	ids = next(iter(ids.batch(NUM_TEST_IMAGES))).numpy().astype('U')
	np.savetxt(output_filename, np.rec.fromarrays([ids, predictions]), fmt = ['%s', '%d'], delimiter = ',', header = 'id,label', comments = '')
	print(f"predictions obtained with filename as {output_filename}")

## function to obtain predictions directly from the dataset
def get_predictions(testTFdataset, model, output_filename):
	import numpy as np
	images = testTFdataset.map(lambda image, idnum : image)
	test_ids = testTFdataset.map(lambda image, idnum : idnum).unbatch()

	NUM_TEST_IMAGES = 0
	for data in testTFdataset.unbatch():
		NUM_TEST_IMAGES += 1

	#probabilities = model.predict(images)
	#predictions = np.argmax(probabilities, axis = -1)
	predictions = model.predict_classes(images)
	predictions = np.reshape(predictions, newshape = (NUM_TEST_IMAGES,))
	print("Generating Output File..")
	test_ids = next(iter(test_ids.batch(NUM_TEST_IMAGES))).numpy().astype('U')
	np.savetxt(output_filename, np.rec.fromarrays([test_ids, predictions]), fmt=['%s', '%d'], delimiter=',', header='id,label', comments='')
	print(f"predictions obtained with filename as -> {output_filename}")


	
## functions for ensembling using Models
def ensemble_model_average(models_list, testTFdataset, classification_type):
	import numpy as np

	test_images_ds = testTFdataset.map(lambda image, idnum : image)
	NUM_IMAGES = 0
	for data in test_images_ds.unbatch():
		NUM_IMAGES += 1
	if classification_type == 'multiclass':
		probs = np.average([models_list[i].predict(test_images_ds) for i in range(len(models_list))], axis = 0)
		preds = np.argmax(probs, axis = -1)
	elif classification_type == 'binary':
		
		preds = np.average([models_list[i].predict(test_images_ds) for i in range(len(models_list))], axis = 0)
		preds = np.array([1 if i >= 0.5 else 0 for i in preds])
		preds = np.reshape(preds, newshape = (NUM_IMAGES,) )
	else:
		print("Choose classification type between 'multiclass' & 'binary'...Not a valid Input.. Exiting")
		return
	return preds

def ensemble_model_weighted(weights, models_list, testTFdataset, classification_type):
	import numpy as np
	probs = []
	test_images_ds = testTFdataset.map(lambda image, idnum : image)
	NUM_IMAGES = 0
	for data in test_images_ds.unbatch():
		NUM_IMAGES += 1
	
	if classification_type == 'binary':
		preds = np.array(sum([models_list[i].predict(test_images_ds) * weights[i] for i in range(len(models_list))]))    
		preds = np.array([1 if i >= 0.5 else 0 for i in preds])
		preds = np.reshape(preds, newshape = (NUM_IMAGES,))
		return preds
	elif classification_type == 'multiclass':
		
		for model in models_list:
			probs.append(model.predict(test_images_ds))

		final_probs = None
		for i, w in enumerate(weights):
			final_probs = w * probs[i]

		preds = np.argmax(final_probs, axis = -1)
	
	else:
		print("Classification type not between 'binary' and 'multiclass'. Please choose the correct classification type...Exiting..")
		return
	return preds


def ensemble_predictions(models_list,  testTFdataset, ensemble_type = 'Model Averaging', classification_type = 'mutliclass', weights = None):
	import numpy as np
	if ensemble_type == 'Model Averaging':
		print("Computing predictions through Model Averaging")
		preds =  ensemble_model_average(models_list, testTFdataset, classification_type)
		print("Generating output file...")

		ids = testTFdataset.map(lambda image, idnum : idnum).unbatch()

		NUM_TEST_IMAGES = 0
		for data in testTFdataset.unbatch():
			NUM_TEST_IMAGES += 1

		ids = next(iter(ids.batch(NUM_TEST_IMAGES))).numpy().astype('U')
		np.savetxt('ensemble_model_averaging.csv', np.rec.fromarrays([ids, preds]), fmt = ['%s', '%d'], delimiter = ',', header = 'id,label', comments = '')
		print("Predictions obtained with the output filename as -> ensemble_model_average")

	elif ensemble_type == 'Model Weighted':
		if weights is None:
			print("Weights not defined for Model Weighted..Exiting...\n Please define weights in a list such that sum(weights) = 1.")
			return
			
		if len(weights) != len(models_list):
			print("Length of weights doesn't match length of models_list... Exiting...")
			return 
		print("Computing predictions through Model Weighted")
		preds =  ensemble_model_weighted(weights, models_list, testTFdataset, classification_type)

		print("Generating output file...")

		ids = testTFdataset.map(lambda image, idnum : idnum).unbatch()

		NUM_TEST_IMAGES = 0
		for data in testTFdataset.unbatch():
			NUM_TEST_IMAGES += 1

		ids = next(iter(ids.batch(NUM_TEST_IMAGES))).numpy().astype('U')
		np.savetxt('ensemble_model_weighted.csv', np.rec.fromarrays([ids, preds]), fmt = ['%s', '%d'], delimiter = ',', header = 'id,label', comments = '')
		print("Predictions obtained with the output filename as -> ensemble_model_weighted.csv")


	else:
		print("Please choose ensemble_type between 'Model Averaging' or 'Model Weighted'")
		return
	#return preds



## function for having test time augmentations
def test_time_augmentations():
	print("Feature Coming Soon!")


if __name__ != '__main__':
	import os
	#os.system('pip install tensorflow==2.2.0')
	import tensorflow as tf 
	import numpy as np
	if tf.__version__ != '2.4.0':
		print("Tensorflow version mismatch. Either pip install tf version as specified in the docs or import the libraries in the beginning.")



def get_models_training_report(models,tpu, n_class, traindata, steps_per_epoch, epochs, batch_size, val_data,  classification_model = 'default', freeze = False, input_shape = [512,512,3], activation = 'softmax', weights = "imagenet", optimizer = "adam", loss = "sparse_categorical_crossentropy", metrics = "sparse_categorical_accuracy", callbacks = None, plot = False):
	
	import pandas as pd
	import numpy as np
	import gc
	from quick_ml.load_models_quick import create_model
	print("Num classes => " , n_class)
	print("\n\n", '#'*90, '\n\n')
	print("\n\n TO OBTAIN THE BEST PERFORMANCE, PLEASE USE THE PRETRAINED MODEL WEIGHTS AS INPUT. Dataset Link -> https://www.kaggle.com/superficiallybot/tf-keras-pretrained-weights\n\n")
	print("\n\n", '#'*90, '\n\n')
	
	df = pd.DataFrame(columns = ['Model_Name', 'Accuracy_top1', 'Accuracy_top3', "Val_Accuracy_top1", "Val_Accuracy_top3"])

	for m in models:
		print(f"Beginning with model -> {m}")
		tf.tpu.experimental.initialize_tpu_system(tpu)
		strategy = tf.distribute.experimental.TPUStrategy(tpu)
		with strategy.scope():
			if classification_model != 'default':
				model = create_model( freeze = freeze, input_shape = input_shape, activation = activation, weights = weights, optimizer = optimizer, loss = loss, metrics = metrics, classes = n_class, model_name = m, classification_model = classification_model)
			else:
				model = create_model( freeze = freeze, input_shape = input_shape, activation = activation, weights = weights, optimizer = optimizer, loss = loss, metrics = metrics , classes = n_class, model_name = m)
				

		history = model.fit(traindata, steps_per_epoch = steps_per_epoch, epochs = epochs,batch_size = batch_size, validation_data =  val_data, callbacks = callbacks, verbose = 0)
		tf.keras.backend.clear_session()
		tf.compat.v1.reset_default_graph()
		del model
		gc.collect()
		
		df = df.append(pd.DataFrame([[m, history.history[metrics][-1], np.mean(history.history[metrics][-3:]) , history.history['val_' + metrics][-1], np.mean(history.history['val_' + metrics][-3:])]], columns = ['Model_Name', 'Accuracy_top1', 'Accuracy_top3', "Val_Accuracy_top1", "Val_Accuracy_top3"]), ignore_index = True)
		print(f"Done with model -> {m}")

	if plot:

		print("Plotting Feature Coming soon...")

		## under making
		"""import matplotlib.plt as plt
		import seaborn as sns

		sns.lineplot(x = list(range(1,epochs + 1)), y = histories[0].history[metrics], label = 'Training Accuracy');
		sns.lineplot(x = list(range(1,epochs + 1)), y = histories[0].history['val_' + metrics], label = 'Validation Accuracy').set_title(f'{metrics} Plot vs Epoch');
		plt.show()"""

	return df


if __name__ != "__main__":
	#import os
	#print("Installing the reqd libraries...\n")
	#os.system("pip install tensorflow==2.2.0")
	#os.system("pip install keras==2.4.3")
	import tensorflow as tf
	if tf.__version__ != '2.4.0':
		print("Error! Tensorflow version Mismatch...")



def rampup_lrfn(epoch):

	"""
	Fine Tune the pretrained model weights in a better and smooth manner.
	Use this as a callback feature during the model training.
	"""
	
	if epoch < LR_RAMPUP_EPOCHS:
		lr = (LR_MAX - LR_START) / LR_RAMPUP_EPOCHS * epoch + LR_START
	elif epoch < LR_RAMPUP_EPOCHS + LR_SUSTAIN_EPOCHS : 
		lr = LR_MAX
	else:
		lr = (LR_MAX - LR_MIN) + LR_EXP_DECAY**(epoch - LR_RAMPUP_EPOCHS - LR_SUSTAIN_EPOCHS) + LR_MIN

	return lr

def stepped_lrfn(epoch):
	return 1e-4 * (0.75 ** np.floor(epoch / 2))

def step_decay(epoch):
	initAlpha = 0.01
	factor = 0.25
	dropEvery = 5
	
	alpha = initAlpha * (factor ** np.floor((1+epoch) / dropEvery))
	return float(alpha)

def simple_lrfn(epoch):

	if epoch < 6:
		return 0.001
	elif 6 < epoch < 12 : 
		return 0.0005
	elif 12 < epoch < 18 : 
		return 0.00008
	else:
		return 0.000006

	return

def plot_lrfn(lr_type = 'rampup'):

	EPOCHS = 20
	import matplotlib.pyplot as plt
	rng = [ i for i in range(EPOCHS)]	
	if lr_type == 'rampup':
		y = [rampup_lrfn(x) for x in range(EPOCHS)]
	elif lr_type == 'simple':
		y = [simple_lrfn(x) for x in range(EPOCHS)]
	elif lr_type == 'stepped':
		y = [stepped_lrfn(x) for x in range(EPOCHS)]
	elif lr_type == 'step_decay':
		y = [step_decay(x) for x in range(EPOCHS)]
	else:
		print("lr_type not among known LRs. If it's a valid LR scheduler method, drop a mail @ antoreepjana@gmail.com and check the developments section of the documentation of quick_ml. https://gitlab.com/antoreep_jana/quick_ml/-/blob/master/README.md")
		return 
		
	print("Learning Rate Schedule => ")
	plt.plot(rng, y); plt.show()
	

def get_callbacks(lr_scheduler = None, early_stopping = None, reduce_lr_on_plateau = None):
	callbacks = []

	#### Learning Rate Scheduler

	if lr_scheduler == 'simple':
		lr_callback = tf.keras.callbacks.LearningRateScheduler(simple_lrfn, verbose = False)

	elif lr_scheduler == 'rampup':
		lr_callback = tf.keras.callbacks.LearningRateScheduler(rampup_lrfn, verbose = False)
	
	elif lr_scheduler == 'stepped':
		lr_callback = tf.keras.callbacks.LearningRateScheduler(stepped_lrfn, verbose = False)
	elif lr_scheduler == 'step_decay':
		lr_callback = tf.keras.callbacks.LearningRateScheduler(step_decay, verbose = False)
	# more callbacks in upcoming versions...
	if lr_callback is not None:
		callbacks.append(lr_callback)

	####  Early Stopping	
	if early_stopping == 'default':
		es_callback = tf.keras.callbacks.EarlyStopping(
				monitor = 'val_loss',
				min_delta = 0,
				patience = 0,
				verbose = 0, 
				mode = 'auto',
				baseline = None,
				restore_best_weights = False
			)
	elif early_stopping is None:
		es_callback = None
	elif isinstance(early_stopping, dict):
		es_callback = tf.keras.callbacks.EarlyStopping(**early_stopping)

	if es_callback is not None:
		callbacks.append(es_callback)



	#### Reduce LR On Plateau
	if reduce_lr_on_plateau == 'default':
		rlop_callback = tf.keras.callbacks.ReduceLROnPlateau(
			monitor = 'val_loss', 
			factor = 0.1, 
			patience = 10, 
			mode = 'auto',
			min_delta = 0.0001, 
			cooldown = 0,
			min_lr = 0)
	elif reduce_lr_on_plateau is None:
		rlop_callback = None
	elif isinstance(reduce_lr_on_plateau, dict):
		rlop_callback = tf.keras.callbacks.ReduceLROnPlateau(**reduce_lr_on_plateau)

	if rlop_callback is not None:
		callbacks.append(rlop_callback)

	return callbacks
		

if __name__ != "__main__":
	global LR_MAX
	global LR_MIN
	global LR_RAMPUP_EPOCHS
	global LR_SUSTAIN_EPOCHS
	global LR_EXP_DEC
	global LR_START
	
	import tensorflow as tf
	if tf.__version__ == '2.4.0':
		pass
	else:
		print("Error! Tensorflow version mismatch...")
	import numpy as np
	LR_MAX = 0.00005 * 8
	LR_MIN = 0.00001
	LR_RAMPUP_EPOCHS = 5
	LR_SUSTAIN_EPOCHS = 0
	LR_EXP_DECAY = 0.8
	LR_START = 0.00001



# perform various data augmentations 


## callbacks supported



def rampup_lrfn(epoch):

	"""
	Fine Tune the pretrained model weights in a better and smooth manner.
	Use this as a callback feature during the model training.
	"""
	
	if epoch < LR_RAMPUP_EPOCHS:
		lr = (LR_MAX - LR_START) / LR_RAMPUP_EPOCHS * epoch + LR_START
	elif epoch < LR_RAMPUP_EPOCHS + LR_SUSTAIN_EPOCHS : 
		lr = LR_MAX
	else:
		lr = (LR_MAX - LR_MIN) + LR_EXP_DECAY**(epoch - LR_RAMPUP_EPOCHS - LR_SUSTAIN_EPOCHS) + LR_MIN

	return lr



def stepped_lrfn(epoch):
	return 1e-4 * (0.75 ** np.floor(epoch / 2))

def step_decay(epoch):
	initAlpha = 0.01
	factor = 0.25
	dropEvery = 5
	
	alpha = initAlpha * (factor ** np.floor((1+epoch) / dropEvery))
	return float(alpha)

def simple_lrfn(epoch):

	if epoch < 6:
		return 0.001
	elif 6 < epoch < 12 : 
		return 0.0005
	elif 12 < epoch < 18 : 
		return 0.00008
	else:
		return 0.000006

	return



def define_callbacks(lr_scheduler = None):
	
	global callbacks
	
	callbacks = {}
	if lr_scheduler is not None:
		callbacks['lr_scheduler'] = lr_scheduler
	else:
		callbacks['lr_scheduler'] = None

	if len(list(callbacks.keys())) == 0:
		print("Error! No callbacks defined but callback called")

## Worked, integrate later
def check_arguments(func):
	flag = True
	def inner(*args, **kwargs):
		
		for arg in args:
			print(arg)
		
		for arg in kwargs:
			print(arg)
					
		if flag:
			func(*args, **kwargs)	
		else:
			print("Error! Check your arguments format! Define the arguments as per the instructions..")
	
	return inner


#@check_arguments
def define_augmentations(flip_left_right = False, hue = None, contrast = None, brightness = None, random_crop = None, random_saturation = None,random_zoom = None, flip_up_down = False, random_rotation = None, random_shear = None, random_shift = None):
	
	
	global augments
	
	augments = {}

	if flip_left_right:
		augments['flip_left_right'] = True

	if hue:
		augments['random_hue'] = hue

	if contrast:
		augments['random_contrast'] = contrast

	if brightness:
		augments['brightness'] = brightness
	
	if random_crop:
		augments['random_crop'] = random_crop
	
	if random_saturation:
		augments['random_saturation'] = random_saturation
	
	if random_zoom:
		augments['random_zoom'] = random_zoom
		
	if flip_up_down:
		augments['flip_up_down'] = True
	
	if random_rotation:
		augments['random_rotation'] = random_rotation
			
	if random_shear:
		augments['random_shear'] = random_shear
	
	if random_shift:
		augments['random_shift'] = random_shift

	print("Defined augmentes -> \n", augments, "\n")
	if len(list(augments.items())) == 0:
		print("Error! Define augments called but no augments defined...")
		return



def augmentations(image, label):

	global augments
	
	for aug in augments.items():
		if aug[0] == 'flip_left_right' and aug[1] == True:
			image = tf.image.random_flip_left_right(image)
		if aug[0] == 'random_hue':
			image = tf.image.random_hue(image, aug[1])
		if aug[0] == 'random_contrast':
			image = tf.image.random_contrast(image, aug[1][0], aug[1][1])
		if aug[0] == 'brightness':
			image = tf.image.random_brightness(image, aug[1])
		if aug[0] == 'random_crop':
			image = tf.image.random_crop(image, [aug[1][0], aug[1][1],aug[1][2]])
		if aug[0] == 'random_saturation':
			image = tf.image.random_saturation(image, aug[1][0], aug[1][1])
		if aug[0] == 'random_zoom':
			image = tf.keras.preprocessing.image.random_zoom(image, (aug[1][0], aug[1][1]))
		if aug[0] == 'flip_up_down' and aug[1] == True:
			image = tf.image.random_flip_up_down(image)
		if aug[0] == 'random_rotation':
			image = tf.keras.preprocessing.image.random_rotation(image, aug[1])
		if aug[0] == 'random_shear':
			image = tf.keras.preprocessing.image.random_shear(image, aug[1])
		if aug[0] == 'random_shift':
			image = tf.keras.preprocessing.image.random_shift(image, aug[1][0], aug[1][1])
		

	return image, label




def get_models_training_report(models,tpu, n_class, GCS_DS_PATH, train_tfrec_path, steps_per_epoch, epochs, batch_size, val_tfrec_path ,  classification_model = 'default', freeze = False, input_shape = [512,512,3], activation = 'softmax', weights = "imagenet", optimizer = "adam", loss = "sparse_categorical_crossentropy", metrics = "sparse_categorical_accuracy", plot = False):
	
	import pandas as pd
	import numpy as np
	import gc
	from quick_ml.load_models_quick import create_model
	from quick_ml.begin_tpu import get_training_dataset
	from quick_ml.begin_tpu import get_validation_dataset
	
	print("Num classes => " , n_class)
	print("\n\n", '#'*90, '\n\n')
	print("\n\n TO OBTAIN THE BEST PERFORMANCE, PLEASE USE THE PRETRAINED MODEL WEIGHTS AS INPUT. Dataset Link -> https://www.kaggle.com/superficiallybot/tf-keras-pretrained-weights\n\n")
	print("\n\n", '#'*90, '\n\n')
	
	df = pd.DataFrame(columns = ['Model_Name', 'Accuracy_top1', 'Accuracy_top3', "Val_Accuracy_top1", "Val_Accuracy_top3"])

	for m in models:
		overall = {}
		print(f"Beginning with model -> {m}")
		tf.tpu.experimental.initialize_tpu_system(tpu)
		strategy = tf.distribute.experimental.TPUStrategy(tpu)
		with strategy.scope():
			if classification_model != 'default':
				model = create_model( freeze = freeze, input_shape = input_shape, activation = activation, weights = weights, optimizer = optimizer, loss = loss, metrics = metrics, classes = n_class, model_name = m, classification_model = classification_model)
			else:
				model = create_model( freeze = freeze, input_shape = input_shape, activation = activation, weights = weights, optimizer = optimizer, loss = loss, metrics = metrics , classes = n_class, model_name = m)
				
		for epoch in range(epochs):
			if epoch == 0:
				#print(f"\nEpoch # {epoch + 1}\n")
				#print(f"\nCurrent Learning Rate -> {K.eval(model.optimizer.lr)}")
				history = model.fit(get_training_dataset(GCS_DS_PATH, train_tfrec_path, batch_size, augmentation = True), validation_data = get_validation_dataset(GCS_DS_PATH, val_tfrec_path, batch_size, augmentation = True), epochs = 1, steps_per_epoch = steps_per_epoch, batch_size = batch_size, verbose = 0)
				if callbacks is not None:
					set_learning_rate(model, epoch + 1)
				
				for k,v in history.history.items():
					overall[k] = v
			
			else:
				#print(f"\nEpoch # {epoch + 1}\n")
				#print(f"\nCurrent Learning Rate -> {K.eval(model.optimizer.lr)}")
				
				history = model.fit(get_training_dataset(GCS_DS_PATH, train_tfrec_path, batch_size, augmentation = True), validation_data = get_validation_dataset(GCS_DS_PATH, val_tfrec_path, batch_size, augmentation = True), epochs = 1, steps_per_epoch = steps_per_epoch, batch_size = batch_size, verbose = 0)
				
				if callbacks is not None:
					set_learning_rate(model, epoch + 1)
			
				for k,v in history.history.items():
					overall[k].extend(v)
				
				
				
		tf.keras.backend.clear_session()
		tf.compat.v1.reset_default_graph()
		del model
		gc.collect()
		
		df = df.append(pd.DataFrame([[m, overall[metrics][-1], np.mean(overall[metrics][-3:]) , overall['val_' + metrics][-1], np.mean(overall['val_' + metrics][-3:])]], columns = ['Model_Name', 'Accuracy_top1', 'Accuracy_top3', "Val_Accuracy_top1", "Val_Accuracy_top3"]), ignore_index = True)
		print(f"Done with model -> {m}")

	if plot:

		print("Plotting Feature Coming soon...")

		## under making
		"""import matplotlib.plt as plt
		import seaborn as sns

		sns.lineplot(x = list(range(1,epochs + 1)), y = histories[0].history[metrics], label = 'Training Accuracy');
		sns.lineplot(x = list(range(1,epochs + 1)), y = histories[0].history['val_' + metrics], label = 'Validation Accuracy').set_title(f'{metrics} Plot vs Epoch');
		plt.show()"""

	return df





def set_learning_rate(model, epoch):
	
	global callbacks
	
	if callbacks['lr_scheduler'] is not None:
		lr_scheduler = callbacks['lr_scheduler']
		if lr_scheduler == 'rampup':
			lr = rampup_lrfn(epoch)
			K.set_value(model.optimizer.learning_rate, lr)
			print(f"Learning Rate changed to {K.eval(model.optimizer.lr)}")
		elif lr_scheduler == 'simple':
			lr = simple_lrfn(epoch)
			K.set_value(model.optimizer.learning_rate, lr)
			print(f"Learning Rate changed to {K.eval(model.optimizer.lr)}")
		elif lr_scheduler == 'step_decay':
			lr = step_decay(epoch)
			K.set_value(model.optimizer.learning_rate, lr)
			print(f"Learning Rate changed to {K.eval(model.optimizer.lr)}")
		elif lr_scheduler == 'stepped':
			lr = step_decay(epoch)
			K.set_value(model.optimizer.learning_rate, lr)
			print(f"Learning Rate changed to {K.eval(model.optimizer.lr)}")
		
	


def augment_and_train(model, GCS_DS_PATH, train_tfrec_path, val_tfrec_path, batch_size, epochs, steps_per_epoch, plot = False):

	from quick_ml.begin_tpu import get_training_dataset
	from quick_ml.begin_tpu import get_validation_dataset

	overall = {}
	
	
	
	for epoch in range(epochs):
	
		if epoch == 0:
			print(f"\nEpoch # {epoch + 1}\n")
			print(f"\nCurrent Learning Rate -> {K.eval(model.optimizer.lr)}")
			history = model.fit(get_training_dataset(GCS_DS_PATH, train_tfrec_path, batch_size, augmentation = True), validation_data = get_validation_dataset(GCS_DS_PATH, val_tfrec_path, batch_size, augmentation = True), epochs = 1, steps_per_epoch = steps_per_epoch, batch_size = batch_size, verbose = 1)
			if callbacks is not None:
				set_learning_rate(model, epoch + 1)
			
			for k,v in history.history.items():
				overall[k] = v
				
				
		else:
		
			print(f"\nEpoch # {epoch + 1}\n")
			print(f"\nCurrent Learning Rate -> {K.eval(model.optimizer.lr)}")
			history = model.fit(get_training_dataset(GCS_DS_PATH, train_tfrec_path, batch_size, augmentation = True), validation_data = get_validation_dataset(GCS_DS_PATH, val_tfrec_path, batch_size, augmentation = True), epochs = 1, steps_per_epoch = steps_per_epoch, batch_size = batch_size, verbose = 1)
			if callbacks is not None:
				set_learning_rate(model, epoch + 1)
			
			for k,v in history.history.items():
				overall[k].extend(v)
		
	
	if plot:
		import matplotlib.pyplot as plt
		import seaborn as sns

		keys = list(overall.keys())
		
		

		sns.lineplot(x = list(range(1, epochs + 1)), y = overall[keys[0]], label = 'Training Accuracy', color = 'k');
		sns.lineplot(x = list(range(1, epochs + 1)), y = overall[keys[2]], label = 'Validation Accuracy', color = 'red');
		plt.show()


	




if __name__ != "__main__":
	global LR_MAX
	global LR_MIN
	global LR_RAMPUP_EPOCHS
	global LR_SUSTAIN_EPOCHS
	global LR_EXP_DEC
	global LR_START
	
	import tensorflow as tf
	if tf.__version__ == '2.4.0':
		pass
	else:
		print("Error! Tensorflow version mismatch...")
	import numpy as np
	LR_MAX = 0.00005 * 8
	LR_MIN = 0.00001
	LR_RAMPUP_EPOCHS = 5
	LR_SUSTAIN_EPOCHS = 0
	LR_EXP_DECAY = 0.8
	LR_START = 0.00001
	import tensorflow.keras.backend as K



def plot_a_grid(columns, rows, img_batch, titles, figsize = (8,8)):
	import matplotlib.pyplot as plt
	try:
		fig=plt.figure(figsize=figsize)
		columns = columns
		rows = rows
		for i in range(1, columns*rows +1):
			img = img_batch[i-1]
			fig.add_subplot(rows, columns, i).title.set_text(titles[i-1])
			plt.imshow(img)
		plt.show()
	except IndexError as err:
		print("ERROR!")
		print(err.args)
		print("The value for rows & columns set wrong. Please ensure columns * rows must be equal to number of images.")

def check_one_image_and_label(tfrecord_filename):
	from quick_ml.begin_tpu import load_dataset
	labeleddata = load_dataset(tfrecord_filename, labeled = True)
	img = None
	label = None
	for data in labeleddata.take(1):
		img = data[0]
		label = data[1]

	import matplotlib.pyplot as plt 
	plt.imshow(img.numpy()); plt.show()

	print("Label of the Image -> ", label.numpy())

def check_one_image_and_label_dataset(labeleddata):
	img = None
	label = None
	for data in labeleddata.take(1):
		img = data[0]
		label = data[1]

	import matplotlib.pyplot as plt 
	plt.imshow(img.numpy()); plt.show()

	print("Label of the Image -> ", label.numpy())
	

def check_batch_and_labels(tfrecord_filename, n_examples, grid_rows, grid_columns, grid_size = (8,8)):
	from quick_ml.begin_tpu import load_dataset
	labeleddata = load_dataset(tfrecord_filename, labeled = True)
	
	img_batch = []
	labels = []
	for data in labeleddata.take(n_examples):
		img_batch.append(data[0].numpy())
		labels.append(data[1].numpy())
	plot_a_grid(grid_columns, grid_rows, img_batch, labels,  grid_size)


def check_batch_and_labels_dataset(labeleddata, n_examples, grid_rows, grid_columns, grid_size = (8,8)):
	img_batch = []
	labels = []
	for data in labeleddata.take(n_examples):
		img_batch.append(data[0].numpy())
		labels.append(data[1].numpy())
	plot_a_grid(grid_columns, grid_rows, img_batch, labels,  grid_size)

def check_one_image_and_id(tfrecord_filename):
	from quick_ml.begin_tpu import load_dataset
	import matplotlib.pyplot as plt

	unlabeleddata = load_dataset(tfrecord_filename, labeled = False)
	img = None
	id_ = None
	for data in unlabeleddata.take(1):
		img = data[0]
		id_ = data[1]

	plt.imshow(img.numpy()); plt.show()
	print("Id of this image is -> ", id_.numpy().decode())


def check_one_image_and_id_dataset(unlabeleddata):
	unlabeleddata = load_dataset(tfrecord_filename, labeled = False)
	img = None
	id_ = None
	for data in unlabeleddata.take(1):
		img = data[0]
		id_ = data[1]

	plt.imshow(img.numpy()); plt.show()
	print("Id of this image is -> ", id_.numpy().decode())


def check_batch_and_ids(tfrecord_filename, n_examples, grid_rows, grid_columns, grid_size = (8,8)):
	
	from quick_ml.begin_tpu import load_dataset
	unlabeleddata = load_dataset(tfrecord_filename, labeled = False)

	img_batch = []
	ids = []
	for data in unlabeleddata.take(n_examples):
		img_batch.append(data[0].numpy())
		ids.append(data[1].numpy().decode())
	plot_a_grid(grid_columns, grid_rows, img_batch, ids, grid_size)
	
	
def check_batch_and_ids_dataset(unlabeleddata, n_examples, grid_rows, grid_columns, grid_size = (8,8)):
	img_batch = []
	ids = []
	for data in unlabeleddata.take(n_examples):
		img_batch.append(data[0].numpy())
		ids.append(data[1].numpy().decode())
	plot_a_grid(grid_columns, grid_rows, img_batch, ids, grid_size)


def create_model(classes, model_name = 'VGG16', classification_model = 'default', freeze = False, input_shape = [512, 512,3], activation  = 'softmax', weights= "imagenet", optimizer = 'adam', loss = 'sparse_categorical_crossentropy', metrics = 'sparse_categorical_accuracy'):
	"""
	
	MODELS -> 'VGG16', 'VGG19',  
	'Xception',
	'DenseNet121', 'DenseNet169', 'DenseNet201', 
	'ResNet50', 'ResNet101', 'ResNet152', 'ResNet50V2', 'ResNet101V2', 'ResNet152V2', 
	'MobileNet', 'MobileNetV2',
	'InceptionV3', 'InceptionResNetV2', 
	'EfficientNetB0', 'EfficientNetB1', 'EfficientNetB2', 'EfficientNetB3', 'EfficientNetB4', 'EfficientNetB5', 'EfficientNetB6', 'EfficientNetB7'

	How to use ->
	
	1) For Effnet & Other than Effnet Models
	   a) with default classification model

	  from load_models_quick import create_model
	  model = create_model(classes = n_classes, model_name = "VGG19")
	   
	   
	   b) With Custom Classification Model
	   
	   from load_models_quick import create_model 
	   class_model = tf.keras.Sequential([
		  tf.keras.lkeras.layers.GlobalAveragePooling2D(),
		  tf.keras.layers.Dense(n_classes, activation = 'softmax') ])
	   model = create_model(classes = n_classes, model_name = 'VGG19', classification_model = 
		class_model)
		
		
	Load multiple models in a go. 
	
	model_names = ['VGG16', 'InceptionV3', 'DenseNet201', "EfficientNetB7"]
	models = []
	for model in model_names:
		models.append(create_model(classes = n_classes, model_name = model))
	
	
	"""
	weights_dict = {
	'DenseNet121' : '/kaggle/input/tf-keras-pretrained-weights/No Top/densenet121_weights_tf_dim_ordering_tf_kernels_notop.h5',
	'DenseNet169' : '/kaggle/input/tf-keras-pretrained-weights/No Top/densenet169_weights_tf_dim_ordering_tf_kernels_notop.h5', 
	'DenseNet201' : '/kaggle/input/tf-keras-pretrained-weights/No Top/densenet201_weights_tf_dim_ordering_tf_kernels_notop.h5',
	'EfficientNetB0' : '/kaggle/input/tf-keras-pretrained-weights/No Top/efficientnet-b0_weights_tf_dim_ordering_tf_kernels_autoaugment_notop.h5',
	'EfficientNetB1' : '/kaggle/input/tf-keras-pretrained-weights/No Top/efficientnet-b1_weights_tf_dim_ordering_tf_kernels_autoaugment_notop.h5',
	'EfficientNetB2' : '/kaggle/input/tf-keras-pretrained-weights/No Top/efficientnet-b2_weights_tf_dim_ordering_tf_kernels_autoaugment_notop.h5',
	'EfficientNetB3' : '/kaggle/input/tf-keras-pretrained-weights/No Top/efficientnet-b3_weights_tf_dim_ordering_tf_kernels_autoaugment_notop.h5',
	'EfficientNetB4' : '/kaggle/input/tf-keras-pretrained-weights/No Top/efficientnet-b4_weights_tf_dim_ordering_tf_kernels_autoaugment_notop.h5',
	'EfficientNetB5' : '/kaggle/input/tf-keras-pretrained-weights/No Top/efficientnet-b5_weights_tf_dim_ordering_tf_kernels_autoaugment_notop.h5',
	'EfficientNetB6' : '/kaggle/input/tf-keras-pretrained-weights/No Top/efficientnet-b6_weights_tf_dim_ordering_tf_kernels_autoaugment_notop.h5' ,
	'EfficientNetB7' : '/kaggle/input/tf-keras-pretrained-weights/No Top/efficientnet-b7_weights_tf_dim_ordering_tf_kernels_autoaugment_notop.h5',
	'InceptionResNetV2' : '/kaggle/input/tf-keras-pretrained-weights/No Top/inception_resnet_v2_weights_tf_dim_ordering_tf_kernels_notop.h5' ,
	'InceptionV3' : '/kaggle/input/tf-keras-pretrained-weights/No Top/inception_v3_weights_tf_dim_ordering_tf_kernels_notop.h5',
	"MobileNet" : '/kaggle/input/tf-keras-pretrained-weights/No Top/mobilenet_1_0_192_tf_no_top.h5',
	"MobileNetV2" : '/kaggle/input/tf-keras-pretrained-weights/No Top/mobilenet_v2_weights_tf_dim_ordering_tf_kernels_1.0_192_no_top.h5',
	"ResNet101" : '/kaggle/input/tf-keras-pretrained-weights/No Top/resnet101_weights_tf_dim_ordering_tf_kernels_notop.h5',
	"ResNet101V2" : '/kaggle/input/tf-keras-pretrained-weights/No Top/resnet101v2_weights_tf_dim_ordering_tf_kernels_notop.h5',
	"ResNet152" : '/kaggle/input/tf-keras-pretrained-weights/No Top/resnet152_weights_tf_dim_ordering_tf_kernels_notop.h5',
	"ResNet152V2" : '/kaggle/input/tf-keras-pretrained-weights/No Top/resnet152v2_weights_tf_dim_ordering_tf_kernels_notop.h5',
	"ResNet50" : '/kaggle/input/tf-keras-pretrained-weights/No Top/resnet50_weights_tf_dim_ordering_tf_kernels_notop.h5' ,
	"ResNet50V2" : '/kaggle/input/tf-keras-pretrained-weights/No Top/resnet50v2_weights_tf_dim_ordering_tf_kernels_notop.h5',
	"VGG16" : '/kaggle/input/tf-keras-pretrained-weights/No Top/vgg16_weights_tf_dim_ordering_tf_kernels_notop.h5',
	"VGG19" : '/kaggle/input/tf-keras-pretrained-weights/No Top/vgg19_weights_tf_dim_ordering_tf_kernels_notop.h5',
	"Xception" : '/kaggle/input/tf-keras-pretrained-weights/No Top/xception_weights_tf_dim_ordering_tf_kernels_notop.h5'
	}
	
	noisy_weights = {
	"EfficientNetB0" : '/kaggle/input/tf-keras-pretrained-weights/No Top/efficientnet-b0_noisy-student_notop.h5',
	"EfficientNetB1" : '/kaggle/input/tf-keras-pretrained-weights/No Top/efficientnet-b1_noisy-student_notop.h5',
	"EfficientNetB2" : '/kaggle/input/tf-keras-pretrained-weights/No Top/efficientnet-b2_noisy-student_notop.h5',
	"EfficientNetB3" : '/kaggle/input/tf-keras-pretrained-weights/No Top/efficientnet-b3_noisy-student_notop.h5',
	"EfficientNetB4" : '/kaggle/input/tf-keras-pretrained-weights/No Top/efficientnet-b4_noisy-student_notop.h5',
	"EfficientNetB5" : '/kaggle/input/tf-keras-pretrained-weights/No Top/efficientnet-b5_noisy-student_notop.h5',
	"EfficientNetB6" : '/kaggle/input/tf-keras-pretrained-weights/No Top/efficientnet-b6_noisy-student_notop.h5',
	"EfficientNetB7" : '/kaggle/input/tf-keras-pretrained-weights/No Top/efficientnet-b7_noisy-student_notop.h5'
	}
	import os
	
	params = {"input_shape" : input_shape, "weights" : weights, "include_top" : False}
	
	weights_dir_present = False
	if os.path.isdir('/kaggle/input/tf-keras-pretrained-weights/No Top'):
		params = {"input_shape" : input_shape, "include_top" : False}
		weights_dir_present = True
	elif os.path.isdir('/content/No Top'):
		params = {'input_shape' : input_shape, 'include_top' : False}
		weights_dir_present = True
		
	
	if classification_model != 'default':
		print("""
		Make Sure that the classification model consists of layers of tf.keras.layers only else it won't work.
			
		""")
	
	import os
	if model_name.startswith("EfficientNet"):
		import os
		os.system("pip install git+https://github.com/qubvel/segmentation_models")
		#import keras
		import tensorflow as tf
	else:
		import tensorflow as tf
		from tensorflow.keras.applications import VGG16, VGG19,  Xception,DenseNet121, DenseNet169, DenseNet201, ResNet50, ResNet101, ResNet152, ResNet50V2, ResNet101V2, ResNet152V2, MobileNet, MobileNetV2, InceptionV3, InceptionResNetV2
		#import keras
	
	
	#tf.keras.backend.clear_session() # error pops
	if model_name == "VGG16":
		if weights_dir_present:
			try:
				pretrained_model = VGG16(weights = weights_dict['VGG16'], **params)
			except:
				pretrained_model = VGG16(weights = os.path.join('/content', '/'.join(weights_dict['VGG16'].split('/')[4:])), **params)
		else:
			pretrained_model = VGG16(**params)
	elif model_name == "VGG19":
		if weights_dir_present:
			
			try:
				pretrained_model = VGG19(weights = weights_dict['VGG19'], **params)
			except:
				pretrained_model = VGG19(weights = os.path.join('/content', '/'.join(weights_dict['VGG19'].split('/')[4:])), **params)
		else:
			pretrained_model = VGG19(**params)
	elif model_name == 'Xception':
		if weights_dir_present:
			try:
				pretrained_model = Xception(weights = weights_dict['Xception'], **params)
			except:
				pretrained_model = Xception(weights = os.path.join('/content', '/'.join(weights_dict['Xception'].split('/')[4:])), **params)
		else:
			pretrained_model = Xception(**params)
	elif model_name == "DenseNet121":
		if weights_dir_present:
			try:
				pretrained_model = DenseNet121(weights = weights_dict['DenseNet121'], **params)
			except:
				pretrained_model = DenseNet121(weights = os.path.join('/content', '/'.join(weights_dict['DenseNet121'].split('/')[4:])), **params)
		else:
			pretrained_model = DenseNet121(**params)
	elif model_name == "DenseNet169":
		if weights_dir_present:
			try:
				pretrained_model = DenseNet169(weights = weights_dict['DenseNet169'], **params)
			except:
				pretrained_model = DenseNet169(weights = os.path.join('/content', '/'.join(weights_dict['DenseNet169'].split('/')[4:])), **params)
		else:
			pretrained_model = DenseNet169(**params)
	elif model_name == "DenseNet201":
		if weights_dir_present:
			try:
				pretrained_model = DenseNet201(weights = weights_dict['DenseNet201'], **params)
			except:
				pretrained_model = DenseNet201(weights = os.path.join('/content', '/'.join(weights_dict['DenseNet201'].split('/')[4:])), **params)
		else:
			pretrained_model = DenseNet201(**params)
	elif model_name == 'EfficientNetB7':
		#import os 
		#os.system("pip install tensorflow==2.1")
		#os.system("pip install keras==2.3.1")
		#os.system("pip install git+https://github.com/qubvel/segmentation_models")
		import efficientnet.tfkeras as efn   
		if weights_dir_present:
				
			try:
				if weights == 'noisy-student':
					pretrained_model = efn.EfficientNetB7(weights = noisy_weights['EfficientNetB7'], **params)
				else:
					pretrained_model = efn.EfficientNetB7(weights = weights_dict['EfficientNetB7'], **params)
			except:
				if weights == 'noisy-student':
					pretrained_model = efn.EfficientNetB7(weights = os.path.join('/content', '/'.join(noisy_weights['EfficientNetB7'].split('/')[4:])), **params)
				else:
					pretrained_model = efn.EfficientNetB7(weights = os.path.join('/content', '/'.join(weights_dict['EfficientNetB7'].split('/')[4:])), **params)
		else:
			pretrained_model = efn.EfficientNetB7(**params)
		
	elif model_name == 'EfficientNetB6':
		
		#os.system("!pip install -q efficientnet")
		#import efficientnet.tfkeras as efficientnet
		#pretrained_model = efficientnet.EfficientNetB6(**params)
		
		import efficientnet.tfkeras as efn  
		if weights_dir_present:
		
			
			try:
			
				if weights == 'noisy-student':
					pretrained_model = efn.EfficientNetB6(weights = noisy_weights['EfficientNetB6'], **params)
				else:
					pretrained_model = efn.EfficientNetB6(weights = weights_dict['EfficientNetB6'], **params)
					
			except:
				
				if weights == 'noisy-student':
					pretrained_model = efn.EfficientNetB6(weights = os.path.join('/content', '/'.join(noisy_weights['EfficientNetB6'].split('/')[4:])), **params)
				else:
					pretrained_model = efn.EfficientNetB6(weights = os.path.join('/content', '/'.join(weights_dict['EfficientNetB6'].split('/')[4:])), **params)
		else:
			pretrained_model = efn.EfficientNetB6(**params)
		
	elif model_name == 'EfficientNetB5':
		#os.system("!pip install -q efficientnet")
		#import efficientnet.tfkeras as efficientnet
		#pretrained_model = efficientnet.EfficientNetB5(**params)
		
		import efficientnet.tfkeras as efn   
		
		if weights_dir_present:
			try:
				
				if weights == 'noisy-student':
					pretrained_model = efn.EfficientNetB5(weights = noisy_weights['EfficientNetB5'], **params)
				else:
					pretrained_model = efn.EfficientNetB5(weights = weights_dict['EfficientNetB5'], **params)
			except:
				if weights == 'noisy-student':
					pretrained_model = efn.EfficientNetB5(weights = os.path.join('/content', '/'.join(noisy_weights['EfficientNetB5'].split('/')[4:])), **params)
				else:
					pretrained_model = efn.EfficientNetB5(weights = os.path.join('/content', '/'.join(weights_dict['EfficientNetB5'].split('/')[4:])), **params)
		else:
			pretrained_model = efn.EfficientNetB5(**params)
		
	elif model_name == 'EfficientNetB4':
		#os.system("!pip install -q efficientnet")
		#import efficientnet.tfkeras as efficientnet
		#pretrained_model = efficientnet.EfficientNetB4(**params)
		
		import efficientnet.tfkeras as efn
		
		if weights_dir_present:
			try:
				if weights == 'noisy-student':
					pretrained_model = efn.EfficientNetB4(weights = noisy_weights['EfficientNetB4'], **params)
				else:
					pretrained_model = efn.EfficientNetB4(weights = weights_dict['EfficientNetB4'], **params)
			except:
				if weights == 'noisy-student':
					pretrained_model = efn.EfficientNetB4(weights = os.path.join('/content', '/'.join(noisy_weights['EfficientNetB4'].split('/')[4:])), **params)
				else:
					pretrained_model = efn.EfficientNetB4(weights = os.path.join('/content', '/'.join(weights_dict['EfficientNetB4'].split('/')[4:])), **params)
		else:
			pretrained_model = efn.EfficientNetB4(**params)
		
	elif model_name == 'EfficientNetB3':
		#os.system("!pip install -q efficientnet")
		#import efficientnet.tfkeras as efficientnet
		#pretrained_model = efficientnet.EfficientNetB3(**params)
		
		import efficientnet.tfkeras as efn  
		
		if weights_dir_present:
			try:
				if weights == 'noisy-student':
					pretrained_model = efn.EfficientNetB3(weights = noisy_weights['EfficientNetB3'], **params)
				else:
					pretrained_model = efn.EfficientNetB3(weights = weights_dict['EfficientNetB3'], **params)
			except:
				if weights == 'noisy-student':
					pretrained_model = efn.EfficientNetB3(weights = os.path.join('/content', '/'.join(noisy_weights['EfficientNetB3'].split('/')[4:])), **params)
				else:
					pretrained_model = efn.EfficientNetB3(weights = os.path.join('/content', '/'.join(weights_dict['EfficientNetB3'].split('/')[4:])), **params)
		else:
			pretrained_model = efn.EfficientNetB3(**params)
		
	elif model_name == 'EfficientNetB2':
		#os.system("!pip install -q efficientnet")
		#import efficientnet.tfkeras as efficientnet
		#pretrained_model = efficientnet.EfficientNetB2(**params)
		
		import efficientnet.tfkeras as efn 
		
		if weights_dir_present:
			try:
				if weights == 'noisy-student':
					pretrained_model = efn.EfficientNetB2(weights = noisy_weights['EfficientNetB2'], **params)
				else:
					pretrained_model = efn.EfficientNetB2(weights = weights_dict['EfficientNetB2'], **params)
			except:
				if weights == 'noisy-student':
					pretrained_model = efn.EfficientNetB2(weights = os.path.join('/content', '/'.join(noisy_weights['EfficientNetB2'].split('/')[4:])), **params)
				else:
					pretrained_model = efn.EfficientNetB2(weights = os.path.join('/content', '/'.join(weights_dict['EfficientNetB2'].split('/')[4:])), **params)
		else:
			pretrained_model = efn.EfficientNetB2(**params)
		
	elif model_name == 'EfficientNetB1':
		#os.system("!pip install -q efficientnet")
		#import efficientnet.tfkeras as efficientnet
		#pretrained_model = efficientnet.EfficientNetB1(**params)
		
		import efficientnet.tfkeras as efn  
		
		if weights_dir_present:
			try:
				if weights == 'noisy-student':
					pretrained_model = efn.EfficientNetB1(weights = noisy_weights['EfficientNetB1'], **params)
				else:
					pretrained_model = efn.EfficientNetB1(weights = weights_dict['EfficientNetB1'], **params)
			except:
				if weights == 'noisy-student':
					pretrained_model = efn.EfficientNetB1(weights = os.path.join('/content', '/'.join(noisy_weights['EfficientNetB1'].split('/')[4:])), **params)
				else:
					pretrained_model = efn.EfficientNetB1(weights = os.path.join('/content', '/'.join(weights_dict['EfficientNetB1'].split('/')[4:])), **params)
		else:
			pretrained_model = efn.EfficientNetB1(**params)
		
	elif model_name == 'EfficientNetB0':
		#os.system("!pip install -q efficientnet")
		#import efficientnet.tfkeras as efficientnet
		#pretrained_model = efficientnet.EfficientNetB0(**params)
		
		import efficientnet.tfkeras as efn 
		
		if weights_dir_present:
			try:
				if weights == 'noisy-student':
					pretrained_model = efn.EfficientNetB0(weights = noisy_weights['EfficientNetB0'], **params)
				else:
					pretrained_model = efn.EfficientNetB0(weights = weights_dict['EfficientNetB0'], **params)
			except:
				if weights == 'noisy-student':
					pretrained_model = efn.EfficientNetB0(weights = os.path.join('/content', '/'.join(noisy_weights['EfficientNetB0'].split('/')[4:])), **params)
				else:
					pretrained_model = efn.EfficientNetB0(weights = os.path.join('/content', '/'.join(weights_dict['EfficientNetB0'].split('/')[4:])), **params)
		else:
			pretrained_model = efn.EfficientNetB0(**params)
		
	elif model_name == "InceptionV3":
		
		if weights_dir_present:
			try:
				pretrained_model = InceptionV3(weights = weights_dict['InceptionV3'], **params)
			except:
				pretrained_model = InceptionV3(weights = os.path.join('/content', '/'.join(weights_dict['InceptionV3'].split('/')[4:])), **params)
		else:
			pretrained_model = InceptionV3(**params)
	elif model_name == "ResNet50":
		if weights_dir_present:
			try:
				pretrained_model = ResNet50(weights = weights_dict['ResNet50'], **params)
			except:
				pretrained_model = ResNet50(weights = os.path.join('/content', '/'.join(weights_dict['ResNet50'].split('/')[4:])), **params)
		else:
			pretrained_model = ResNet50(**params)
	elif model_name == "ResNet101":
		if weights_dir_present:
			try:
				pretrained_model = ResNet101(weights = weights_dict['ResNet101'], **params)
			except:
				pretrained_model = ResNet101(weights = os.path.join('/content', '/'.join(weights_dict['ResNet101'].split('/')[4:])), **params)
		else:
			pretrained_model = ResNet101(**params)
	elif model_name == "ResNet152":
		
		if weights_dir_present:
			try:
				pretrained_model = ResNet152(weights = weights_dict['ResNet152'], **params)
			except:
				pretrained_model = ResNet152(weights = os.path.join('/content', '/'.join(weights_dict['ResNet152'].split('/')[4:])), **params)
		else:
			pretrained_model = ResNet152(**params)
	elif model_name == "ResNet50V2":
		
		if weights_dir_present:
			try:
				pretrained_model = ResNet50V2(weights = weights_dict['ResNet50V2'], **params)
			except:
				pretrained_model = ResNet50V2(weights = os.path.join('/content', '/'.join(weights_dict['ResNet50V2'].split('/')[4:])), **params)
		else:
			pretrained_model = ResNet50V2(**params)
	elif model_name == "ResNet101V2":
		
		if weights_dir_present:
			try:
				pretrained_model = ResNet101V2(weights = weights_dict['ResNet101V2'], **params)
			except:
				pretrained_model = ResNet101V2(weights = os.path.join('/content', '/'.join(weights_dict['ResNet101V2'].split('/')[4:])), **params)
		else:
			pretrained_model = ResNet101V2(**params)
			
	elif model_name == "ResNet152V2":
		
		if weights_dir_present:
			try:
				pretrained_model = ResNet152V2(weights = weights_dict['ResNet152V2'], **params)
			except:
				pretrained_model = ResNet152V2(weights = os.path.join('/content', '/'.join(weights_dict['ResNet152V2'].split('/')[4:])), **params)
		else:
			pretrained_model = ResNet152V2(**params)
	elif model_name == 'MobileNet':
		
		if weights_dir_present:
			try:
				pretrained_model = MobileNet(weights = weights_dict['MobileNet'], **params)
			except:
				pretrained_model = MobileNet(weights = os.path.join('/content', '/'.join(weights_dict['MobileNet'].split('/')[4:])), **params)
		else:
			pretrained_model = MobileNet(**params)    
	elif model_name == 'MobileNetV2':
		
		if weights_dir_present:
			try:
				pretrained_model = MobileNetV2(weights = weights_dict['MobileNetV2'], **params)
			except:
				pretrained_model = MobileNetV2(weights = os.path.join('/content', '/'.join(weights_dict['MobileNetV2'].split('/')[4:])), **params)
		else:
			pretrained_model = MobileNetV2(**params)
			
	elif model_name == "InceptionResNetV2":
		
		if weights_dir_present:
			try:
				pretrained_model = InceptionResNetV2(weights = weights_dict['InceptionResNetV2'], **params)
			except:
				pretrained_model = InceptionResNetV2(weights = os.path.join('/content', '/'.join(weights_dict['InceptionResNetV2'].split('/')[4:])), **params)
		else:
			pretrained_model = InceptionResNetV2(**params)

	else:
		print("model not among known models... exiting...")
		return
	
	if freeze:
		pretrained_model.trainable = False
	else:
		#pretrained_model.trainable = True
		for layer in pretrained_model.layers:
			if not isinstance(layer, tf.keras.layers.BatchNormalization):
				layer.trainable = True
			else:
				layer.trainable = False
	
	if model_name.startswith("EfficientNet"):
		
		if classification_model == 'default':
			#tf.keras.backend.clear_session()
			model = tf.keras.Sequential([
			pretrained_model,
			tf.keras.layers.GlobalAveragePooling2D(),
			tf.keras.layers.Dense(classes, activation = activation)
			])
		else:
			#tf.keras.backend.clear_session()
			new_model = tf.keras.Sequential()
			new_model.add(pretrained_model)
			for layer in classification_model.layers:
				new_model.add(layer)
			new_model.compile(optimizer = optimizer, loss = loss, metrics = [metrics])
			return new_model
	else:
		
		if classification_model == 'default':
			#tf.keras.backend.clear_session()
			model = tf.keras.Sequential([
			pretrained_model,
			tf.keras.layers.GlobalAveragePooling2D(),
			tf.keras.layers.Dense(classes, activation =activation)
			])
		else:
			#tf.keras.backend.clear_session()
			new_model = tf.keras.Sequential()
			new_model.add(pretrained_model)
			for layer in classification_model.layers:
				new_model.add(layer)
			new_model.compile(optimizer = optimizer, loss = loss, metrics = [metrics])
			return new_model
				
	
	model.compile(optimizer = optimizer, loss = loss, metrics = [metrics])
	
	return model    

if __name__ != "__main__":
	#print("Please refer to help(create_model) to know about how to use.\n")
	#import os
	#print("Installing the reqd libraries...\n")
	#os.system("pip install tensorflow==2.2.0")
	#os.system("pip install keras==2.4.3")
	import tensorflow as tf
	if tf.__version__ != '2.4.0':
		print("Error! Tensorflow version mismatch...")



from setuptools import setup, Extension, find_packages

with open('README.md') as f:
	readme = f.read()

setup(
  name = 'quick_ml',         # How you named your package folder (MyLib)
  #packages = ['quick_ml'],   # Chose the same as "name"
  packages = find_packages(),
  version = '1.3.16',      # Start with a small number and increase it with every change you make
  license='MIT',        # Chose a license from here: https://help.github.com/articles/licensing-a-repository
  description = 'quick_ml : ML For Everyone. Official Website -> www.antoreepjana.wixsite.com/quick-ml. Making Deep Learning through TPUs accessible to everyone. Lesser Code, faster computation, better modelling.',   # Give a short description about your library
  long_description = readme,
  long_description_content_type = 'text/markdown',
  author = 'Antoreep Jana',                   # Type in your name
  author_email = 'antoreepjana@gmail.com',      # Type in your E-Mail
  url = 'https://gitlab.com/antoreep_jana/quick_ml',   # Provide either the link to your github or to your website
  download_url = 'https://gitlab.com/antoreep_jana/quick_ml/-/archive/v1.3.16/quick_ml-v1.3.16.tar.gz',    # I explain this later on
  keywords = ['quick_ml','quick ml','TPU', 'Deep Learning TPU', 'tensorflow', 'deep learning'],   # Keywords that define your package best
  install_requires=[            # I get to this in a second
          'tensorflow==2.4.0'
      ],

  classifiers=[
    'Development Status :: 5 - Production/Stable',      # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
    'Intended Audience :: Developers',      # Define that your audience are developers
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: MIT License',   # Again, pick a license
    'Programming Language :: Python :: 3',      #Specify which pyhton versions that you want to support
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
  ],
)



def _int64_feature(value):
	return tf.train.Feature(int64_list = tf.train.Int64List(value = [value]))

def _bytes_feature(value):
	return tf.train.Feature(bytes_list = tf.train.BytesList(value = [value]))

def load_image(addr, IMAGE_SIZE = (192,192)):
	# read the img & resize to the image size
	# converting the loaded BGR image to RGB image

	img = cv2.imread(addr)
	if img is None:
		return None
	img = cv2.resize(img, IMAGE_SIZE, interpolation = cv2.INTER_CUBIC)
	img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
	return img

def get_addrs_labels(data_dir):

	data_path = data_dir + '/*/*.jpg'

	addrs = glob.glob(data_path)
	
	classes = os.listdir(data_dir)
	classes.sort()

	labels_dict = {}
	for i, class_ in enumerate(classes):
		labels_dict[class_] = i
	print(f"Class Encodings for the Dataset Folder is as follows -> \n\n {labels_dict}")
	
	labels = []
	for addr in addrs:
		label = None
		for i in classes:
			if i in addr:
				label = i
				break
		labels.append(labels_dict[label])

	# shuffling the data
	c = list(zip(addrs, labels))
	shuffle(c)

	addrs, labels = zip(*c)

	return addrs, labels


def affirmationcsv(func):
	def inner(*args, **kwargs):

		print("""Please ensure that the data format is of the following structure. -> \n

			csv file format. 

			| Image      | Id    |
			------------------
			|Image1.jpg  | classA|
			|Image2.jpg  | classB|
			|Image3.jpg  | classA|
			|Image4.jpg  | classC|
			|ImageN.jpg  | classN|

			AND

			DATA_DIR of Images follows the following structure -> \n

			/data 
				|
				|-> Image1.jpg
				|-> Image2.jpg
				|-> Image3.jpg
				|
				|-> ImageN.jpg


			""")
		func(*args, **kwargs)

	return inner

@affirmationcsv
def create_tfrecords_from_csv(data_dir, csv_path, outputfilename, IMAGE_SIZE = (192,192)):
	

	import pandas as pd 
	df = pd.read_csv(csv_path)

	df.sort_values(['Image'], axis = 0, inplace = True)

	addrs = glob.glob(data_dir + '/*.jpg')
	addrs.sort()
	labels = list(df['Id'])
	
	labels_dict = {}
	labels_unique = list(set(labels))
	labels_unique.sort()

	for i, class_ in enumerate(labels_unique):
		labels_dict[class_] = i
		
	print(f"Encodings for the classes are -> {labels_dict}")
	
	labels = [labels_dict[i] for i in labels]
	
	c = list(zip(addrs, labels))
	shuffle(c)


	addrs, labels = zip(*c)
	create_tfrecord_labeled(addrs, labels, outputfilename, IMAGE_SIZE)



@affirmationcsv
def create_split_tfrecords_from_csv(data_dir, csv_path, outfile1name, outfile2name, split_size_ratio , IMAGE_SIZE = (192,192)):
	
	import pandas as pd
	df = pd.read_csv(csv_path)

	df.sort_values(['Image'], axis = 0, inplace = True)

	addrs = glob.glob(data_dir + '/*.jpg')
	addrs.sort()
	labels = list(df['Id'])
	
	labels_dict = {}
	labels_unique = list(set(labels))
	labels_unique.sort()

	for i, class_ in enumerate(labels_unique):
		labels_dict[class_] = i
		
	print(f"Encodings for the classes are -> {labels_dict}")
	
	labels = [labels_dict[i] for i in labels]
	
	c = list(zip(addrs, labels))
	shuffle(c)
	addrs, labels = zip(*c)

	file1_addrs = addrs[0 : int(split_size_ratio * len(addrs))]
	file1_labels = labels[0: int(split_size_ratio * len(labels))]

	file2_addrs = addrs[int(split_size_ratio * len(addrs)) :]
	file2_labels =  labels[int(split_size_ratio * len(labels)):]

	create_tfrecord_labeled(file1_addrs, file1_labels, outfile1name, IMAGE_SIZE)
	create_tfrecord_labeled(file2_addrs, file2_labels, outfile2name, IMAGE_SIZE)

def retrieve_file_paths(dirName):
	filePaths = []
	
	for root, directories, files in os.walk(dirName):
		for filename in files:
			filePath = os.path.join(root, filename)
			filePaths.append(filePath)
			
	return filePaths


def create_split_tfrecords_data(data_dir, outfile1name, output1folder, outfile2name, output2folder, split_size_ratio,num_parts1 = 1, num_parts2 = 1,  IMAGE_SIZE = (192,192), zip = False):

	print(f"Split ratio -> {split_size_ratio}")
	addrs, labels = get_addrs_labels(data_dir)

	file1_addrs = addrs[0 : int(split_size_ratio * len(addrs))]
	file1_labels = labels[0: int(split_size_ratio * len(labels))]

	file2_addrs = addrs[int(split_size_ratio * len(addrs)) :]
	file2_labels =  labels[int(split_size_ratio * len(labels)):]
	
	os.mkdir(output1folder); os.mkdir(output2folder)
	
	from pathlib import Path
	os.chdir(output1folder)
	create_tfrecord_labeled(file1_addrs, file1_labels, outfile1name, IMAGE_SIZE, num_parts1)
	os.chdir(Path(os.getcwd()).parent)
	os.chdir(output2folder)
	create_tfrecord_labeled(file2_addrs, file2_labels, outfile2name, IMAGE_SIZE, num_parts2)
	os.chdir(Path(os.getcwd()).parent)
	
	if zip:
		folder1_files = retrieve_file_paths(str(Path(output1folder).absolute()))
		folder2_files = retrieve_file_paths(str(Path(output2folder).absolute()))
		
		import zipfile
		
		zip_file = zipfile.ZipFile('data' + '.zip', 'w')
		with zip_file:
			for file in folder1_files:
				zip_file.write(file)
			for file in folder2_files:
				zip_file.write(file)
	
	print("Data Successfully created!")

def affirmationfolder(func):
	def inner(*args,**kwargs):
		print("""
		Ensure that the data directory has the following structure ->

		/data|
			 |
			 |-> class1
			 |-> class2
			 |-> class3
			 |-> class4
			 .
			 .
			 .
			 |-> classN

		the input to data_dir must the path to /data folder. & Make sure that the images are in .jpg format
		""")
		func(*args, **kwargs)
	return inner

@affirmationfolder
def create_tfrecord_labeled(addrs, labels, out_filename, IMAGE_SIZE = (192,192), num_parts = 1):



	print(f"Beginning to write data to {out_filename}\n\n")
	if num_parts == 1:
		
		writer = tf.io.TFRecordWriter(out_filename)
		for i in range(len(addrs)):
			if i % 300 == 0:
				print("Data written -> {}/{}".format(i, len(addrs)))
				sys.stdout.flush()

			# loading the image
			img = load_image(addrs[i], IMAGE_SIZE)
			label = labels[i]

			if img is None:
				continue

			# feature creation

			feature = {
				'image' : _bytes_feature(img.tostring()),
				'label' : _int64_feature(label)
			}



			# example protocol buffer
			example = tf.train.Example(features = tf.train.Features(feature = feature))

			# serialize the write the output to the file
			writer.write(example.SerializeToString())
		writer.close()
		sys.stdout.flush()
		print(f"Done with writing data to {out_filename}\n\n")
		print("""
				Your Labeled TFRecord Format is -> 

				{
				'image' : tf.io.FixedLenFeature([], tf.string),
				'label' : tf.io.FixedLenFeature([], tf.int64)
				}

				You would be asked the TFRecord Format while reading the training and validation dataset.
		""")
		print("Now after you have obtained the tfrecords, you need to upload them to GCS Buckets to utlize TPU computation...")
		print("To do so, the easiest method is uploading on Kaggle.")
		print("""
			1. Public Datasets

			from kaggle_datasets import KaggleDatasets
			GCS_DS_PATH = KaggleDatasets().get_gcs_path()


			2. Private Datasets

			# Get the credentials from the Cloud SDK
			from kaggle_secrets import UserSecretsClient
			user_secrets = UserSecretsClient()
			user_credential = user_secrets.get_gcloud_credential()

			# Set the credentials
			user_secrets.set_tensorflow_credential(user_credential)

			# Use a familiar call to get the GCS path of the dataset
			from kaggle_datasets import KaggleDatasets
			GCS_DS_PATH = KaggleDatasets().get_gcs_path()
			""")
	else:
		
		chunk_size = len(addrs) // num_parts
		
		indx = 0
		
		for part in range(num_parts):
			writer = tf.io.TFRecordWriter(out_filename.split('.')[0] + f'_part_{part+1}.' + out_filename.split('.')[1])
			addrs_ = addrs[indx : indx + chunk_size]
			labels_ = labels[indx : indx + chunk_size]
			if part == num_parts - 1:
				addrs_ = addrs[indx:]
				labels_ = labels[indx:]
			for i in range(len(addrs_)):
				if i % 100 == 0:
					print("Data written -> {}/{}".format(i, len(addrs_)))
					sys.stdout.flush()

				# loading the image
				img = load_image(addrs_[i], IMAGE_SIZE)
				label = labels_[i]

				if img is None:
					continue

				# feature creation

				feature = {
					'image' : _bytes_feature(img.tostring()),
					'label' : _int64_feature(label)
				}



				# example protocol buffer
				example = tf.train.Example(features = tf.train.Features(feature = feature))

				# serialize the write the output to the file
				writer.write(example.SerializeToString())
			writer.close()
			sys.stdout.flush()
			indx = indx + chunk_size
			print(f"Done with writing data to {out_filename.split('.')[0] + f'_part_{part+1}.' + out_filename.split('.')[1]}\n\n")
			print("""
				Your Labeled TFRecord Format is -> 

				{
				'image' : tf.io.FixedLenFeature([], tf.string),
				'label' : tf.io.FixedLenFeature([], tf.int64)
				}

				You would be asked the TFRecord Format while reading the training and validation dataset.
			""")
			print("Now after you have obtained the tfrecords, you need to upload them to GCS Buckets to utlize TPU computation...")
			print("To do so, the easiest method is uploading on Kaggle.")
			print("""
				1. Public Datasets

				from kaggle_datasets import KaggleDatasets
				GCS_DS_PATH = KaggleDatasets().get_gcs_path()


				2. Private Datasets

				# Get the credentials from the Cloud SDK
				from kaggle_secrets import UserSecretsClient
				user_secrets = UserSecretsClient()
				user_credential = user_secrets.get_gcloud_credential()

				# Set the credentials
				user_secrets.set_tensorflow_credential(user_credential)
	
				# Use a familiar call to get the GCS path of the dataset
				from kaggle_datasets import KaggleDatasets
				GCS_DS_PATH = KaggleDatasets().get_gcs_path()
				""")

def get_addrs_ids(data_dir):
	data_path = data_dir + '/*.jpg'

	addrs = glob.glob(data_path)

	ids = os.listdir(data_dir)


	return addrs, ids


def create_tfrecord_unlabeled(out_filename, addrs, ids, IMAGE_SIZE = (192,192)):
	print("""
		The Data for this should be unlabeled. 

		The folder structure should be

		data |
			 |
			 |->filename1.jpg
			 |->filename2.jpg
			 |->filename3.jpg
			 |->filename4.jpg
			 |.
			 ..
			 ..
			 |->filenameN.jpg


		""")

	print(f"Beginning to write the {out_filename}\n\n")
	writer = tf.io.TFRecordWriter(out_filename)
	for i in range(len(addrs)):
		if i % 300 == 0:
			print("Data Written -> {}/{}".format(i, len(addrs)))
			sys.stdout.flush()

		#loading the image
		img = load_image(addrs[i], IMAGE_SIZE)
		idnum = ids[i]

		if img is None:
			continue

		#feature creation
		feature = {
			'image' : _bytes_feature(img.tostring()),
			'idnum' : _bytes_feature(idnum.encode())
		}

		# example protocol buffer
		example = tf.train.Example(features = tf.train.Features(feature = feature))


		# serialize & write
		writer.write(example.SerializeToString())
	writer.close()
	sys.stdout.flush()

	print("""
			Your Unlabeled TFRecord Format is -> 

			{
			'image' : tf.io.FixedLenFeature([], tf.string),
			'idnum' : tf.io.FixedLenFeature([], tf.string)
			}

			You would be asked the TFRecord Format while reading the test dataset.
	""")
	print("After you have obtained the tfrecords, you need to upload the tfrecords to GCS Buckets for TPU computation")
	print("To do so, the easiest method is to upload on Kaggle. Either as a public or as a private dataset.")
	print("""
		NOTE: Fully Functional and working with Public Datasets
		1. Public Datasets

		from kaggle_datasets import KaggleDatasets
		GCS_DS_PATH = KaggleDatasets().get_gcs_path()


		NOTE : Not tested with Private Datasets
		2. Private Datasets

		# Get the credentials from the Cloud SDK
		from kaggle_secrets import UserSecretsClient
		user_secrets = UserSecretsClient()
		user_credential = user_secrets.get_gcloud_credential()

		# Set the credentials
		user_secrets.set_tensorflow_credential(user_credential)

		# Use a familiar call to get the GCS path of the dataset
		from kaggle_datasets import KaggleDatasets
		GCS_DS_PATH = KaggleDatasets().get_gcs_path()

		""")


if __name__ != "__main__":
	import os
	import tensorflow as tf
	if tf.__version__ != '2.4.0':
		print("Tensorflow version Error!")
	from random import shuffle
	import glob
	import sys
	import cv2
	import numpy as np 
